"""
This code defines some utility functions and constants used across this project.
"""

# Constants.
GIT = "git"

#############
# FUNCTIONS #
#############

def get_yes_no(message: str) -> bool:
    """ Get the user to answer a yes/no question. """
    valid = {"yes": True, "y": True, "no": False, "n": False}
    print(message+" [y/n]")
    answer = input().lower()
    if answer in valid:
        return valid[answer]
    return False



"""
This code tests some of the continuous integration functions.
"""

# Source imports.
from source.continuous_integration import print_encased, run_linter

###########
# TESTING #
###########

def test_print_encased():
    """ Test that the function runs without crashing. """
    print_encased("Some random guff.")
    print_encased("Some random guff.", colour="orange")

def test_run_linter():
    """ Test that the function returns the right value. """
    assert run_linter()



"""
This code tests some of the miscellaneous functions.
"""

# Standard imports.
from unittest.mock import patch

# Source imports.
from source.misc import get_yes_no

###########
# TESTING #
###########

@patch("source.misc.input")
def test_get_yes_no(input_mock):
    """ Test that "y" returns True and "n" returns False. """
    input_mock.return_value = "y"
    assert get_yes_no("message")
    input_mock.return_value = "n"
    assert not get_yes_no("message")



"""
This code defines some useful functions when writing a minimal continuous
integration routine.
"""

# Standard imports.
import shutil
import subprocess
from glob import glob
from pathlib import Path

# Non-standard imports.
import pytest
from termcolor import colored

DEFAULT_PATH_TO_LINTER_RC = "pylintrc"
PATH_TO_BACKUP_LINTER_RC = \
    str(Path(__file__).parent/"backup_configs"/"backup_pylintrc")
DEFAULT_PATH_TO_TEST_INI = "pytest.ini"
PATH_TO_BACKUP_TEST_INI = \
    str(Path(__file__).parent/"backup_configs"/"backup_pytest.ini")
PIP_INSTALL_THIS = ("pip", "install", ".")

#############
# FUNCTIONS #
#############

def print_encased(message, symbol="#", colour=None):
    """ Print the message encased in hashes. """
    message_line = symbol+" "+message+" "+symbol
    hashes = ""
    for _ in range(len(message_line)):
        hashes = hashes+symbol
    if colour:
        print(colored(" ", colour))
        print(colored(hashes, colour))
        print(colored(message_line, colour))
        print(colored(hashes, colour))
        print(colored(" ", colour))
    else:
        print(" ")
        print(hashes)
        print(message_line)
        print(hashes)
        print(" ")

def run_tests(path_to_test_ini=DEFAULT_PATH_TO_TEST_INI):
    """ Run PyTest. """
    if not Path(path_to_test_ini).exists():
        shutil.copy(PATH_TO_BACKUP_TEST_INI, path_to_test_ini)
    return_code = pytest.main()
    if return_code == 0:
        return True
    return False

def run_linter(path_to_linter_rc=DEFAULT_PATH_TO_LINTER_RC):
    """ Run PyLint on this repo. """
    if not Path(path_to_linter_rc).exists():
        shutil.copy(PATH_TO_BACKUP_LINTER_RC, path_to_linter_rc)
    source_file_paths = glob("**/*.py")
    arguments = ["pylint"]+source_file_paths
    try:
        subprocess.run(arguments, check=True)
    except subprocess.CalledProcessError:
        return False
    return True

def run_continuous_integration_no_print(
        lint=True, test=True, stop_on_failure=False
    ):
    """ Execute a minimal continuous integration routine. """
    lint_result, test_result = True, True
    if lint:
        lint_result = run_linter()
        if (not lint_result) and stop_on_failure:
            return False
    if test:
        test_result = run_tests()
        if (not test_result) and stop_on_failure:
            return False
    if lint_result and test_result:
        return True
    return False

def run_continuous_integration(lint=True, test=True, stop_on_failure=False):
    """ Run this file. """
    print_encased("Starting continuous integration routine...")
    result = \
        run_continuous_integration_no_print(
            lint=lint, test=test, stop_on_failure=stop_on_failure
        )
    if result:
        print_encased("Continuous integration: PASS", colour="green")
    else:
        print_encased("Continuous integration: FAIL", colour="red")
    return result



"""
This code defines a class which backs up the so-called royal repos.
"""

# Standard imports.
import logging
import subprocess
from dataclasses import dataclass, field
from pathlib import Path

# Local imports.
from .hmss_config import HMSSConfig

# Local constants.
PATH_TO_LOG = str(Path.home()/"hm_git.log")
LOG_FORMAT = "%(asctime)s | %(levelname)s | %(message)s"

##############
# MAIN CLASS #
##############

def back_up_royal_repos(*args, **kwargs) -> bool:
    """ Compress the class into a function. """
    backup_obj = RoyalReposBackup(*args, **kwargs)
    return backup_obj.back_up_all()

@dataclass
class RoyalReposBackup:
    """ The class in question. """
    human_interface: bool = False
    config: HMSSConfig = None
    logger: logging.Logger = field(init=False, default=None)

    def __post_init__(self):
        if not self.config:
            self._auto_set_config()
        self._auto_set_logger()

    def _auto_set_config(self):
        """ Ronseal. """
        if self.human_interface:
            self.config = HMSSConfig.read_human()
        else:
            self.config = HMSSConfig.read_machine()

    def _auto_set_logger(self):
        """ Construct our logging object. """
        result = logging.getLogger()
        result.setLevel(logging.INFO)
        formatter = logging.Formatter(LOG_FORMAT)
        file_handler = logging.FileHandler(PATH_TO_LOG)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        result.addHandler(file_handler)
        self.logger = result

    def _fetch_royal(self, path_to: str) -> bool:
        """ Run `git fetch` on a given repo. """
        return self._run_git_command("fetch", path_to)

    def _pull_royal(self, path_to: str) -> bool:
        """ Run `git pull` on a given repo. """
        return self._run_git_command("pull", path_to)

    def _run_git_command(self, command: str, git_directory: str) -> bool:
        """ Run a given Git command in a given Git directory. """
        try:
            subprocess.run(["git", command], check=True, cwd=git_directory)
        except subprocess.CalledProcessError as exc:
            self.logger.error(
                "Non-zero exit code running git %s within %s: %s",
                command,
                git_directory,
                format(exc)
            )
            return False
        return True

    def back_up_all(self) -> bool:
        """ Back up ALL royal repos. """
        result = True
        self.logger.info("Backing up royal repos...")
        for repo in self.config.royal_repos:
            if not self.back_up_one(repo):
                result = False
        if result:
            self.logger.info("Backed up royal repos successfully.")
        else:
            self.logger.error("Error backing up royal repos.")
        return result

    def back_up_one(self, repo_name: str) -> bool:
        """ Back up a given INDIVIDUAL royal repo. """
        path_to = str(Path.home()/repo_name)
        if not (
            self._fetch_royal(path_to) and
            self._pull_royal(path_to)
        ):
            self.logger.error("Error backing up: %s", repo_name)
            return False
        return True



"""
This code tests the HMSSConfig class.
"""

# Standard imports.
from pathlib import Path
from unittest.mock import patch

# Source imports.
from source.hmss_config import HMSSConfig, DEFAULT_HMSS_CONFIG

# Local constants.
TEST_PATH_TO_HMSS_CONFIG = "test_hmss_config.json"

###########
# TESTING #
###########

@patch("source.hmss_config.PATH_TO_HMSS_CONFIG", TEST_PATH_TO_HMSS_CONFIG)
def test_hmss_config():
    """ Test that the class works as intended. """
    Path(TEST_PATH_TO_HMSS_CONFIG).unlink(missing_ok=True)
    assert HMSSConfig.read_human() == None
    config_obj = HMSSConfig.read_machine()
    for key, value in DEFAULT_HMSS_CONFIG.items():
        if key != "path_to_wallpaper_file":
            assert value == getattr(config_obj, key)
    Path(TEST_PATH_TO_HMSS_CONFIG).unlink()



"""
This code defines a class which holds the config values for His Majesty's
Software Suite.
"""

# Standard imports.
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Self

# Local constants.
DEFAULT_GIT_ACCOUNT_NAME = "tomhosker"
DEFAULT_CLONE_METHOD = "https"
DEFAULT_GIT_HOST = "github.com"
JSON_INDENT = 4
# Lists.
DEFAULT_ESSENTIAL_APT_PACKAGES = ["git", "gedit-plugins"]
DEFAULT_NON_ESSENTIAL_APT_PACKAGES = ["inkscape", "vlc"]
DEFAULT_ROYAL_REPOS = [
    "celanta_at_the_well_of_life",
    "chancery",
    "chancery_b",
    "hgmj",
    "hoskers_almanack",
    "hosker_utils",
    "lucifer_in_starlight",
    "reading_room",
    "vanilla_web"
]
# Paths.
PATH_OBJ_TO_HERE = Path(__file__).parent
PATH_TO_DEFAULT_WALLPAPER_DIR = str(PATH_OBJ_TO_HERE/"wallpaper")
PATH_TO_HMSS_CONFIG = str(Path.home()/"hmss_config.json")
# Dicts.
DEFAULT_HMSS_CONFIG = {
    "thunderbird_num": None,
    "essential_apt_packages": DEFAULT_ESSENTIAL_APT_PACKAGES,
    "non_essential_apt_packages": DEFAULT_NON_ESSENTIAL_APT_PACKAGES,
    "path_to_wallpaper_file": None,
    "install_chrome": False,
    "royal_repos": DEFAULT_ROYAL_REPOS,
    "clone_method": DEFAULT_CLONE_METHOD,
    "git_host": DEFAULT_GIT_HOST,
    "git_account_name": DEFAULT_GIT_ACCOUNT_NAME
}

##############
# MAIN CLASS #
##############

@dataclass
class HMSSConfig:
    """ The class in question. """
    thunderbird_num: int|None = None
    essential_apt_packages: list[str]|None = None
    non_essential_apt_packages: list[str]|None = None
    path_to_wallpaper_file: str|None = None
    install_chrome: bool|None = False
    royal_repos: list[str]|None = None
    clone_method: str|None = None
    git_host: str|None = None
    git_account_name: str|None = None

    def __post_init__(self):
        if not self.path_to_wallpaper_file:
            self._set_path_to_wallpaper_file()

    def _set_path_to_wallpaper_file(self):
        """ Set this attribute to its fallback value. """
        if self.thunderbird_num:
            filename = f"wallpaper_t{self.thunderbird_num}.png"
        else:
            filename = "default.jpg"
        self.path_to_wallpaper_file = \
            str(Path(PATH_TO_DEFAULT_WALLPAPER_DIR)/filename)

    @classmethod
    def read(cls) -> Self:
        """ Create an instance of this class by reading the config file. """
        with open(PATH_TO_HMSS_CONFIG, "r") as config_file:
            init_dict = json.load(config_file)
        result = cls(**init_dict)
        return result

    @classmethod
    def read_machine(cls) -> Self:
        """ A machine-interface version of the read() method. """
        return cls.read()

    @classmethod
    def read_human(cls) -> Self|None:
        """ A human-interface version of the read() method. """
        if not Path(PATH_TO_HMSS_CONFIG).exists():
            print("Looks like you don't have an HMSS config file.")
            print(f"I'll create one for you now at {PATH_TO_HMSS_CONFIG}")
            cls.write_defaults()
            print("Presently, all configs are set to their default values.")
            print("Open the file, amened as required, and then run me again.")
            return None
        try:
            result = cls.read()
        except TypeError:
            print("Looks like there's something wrong with your config file.")
            print(f"Path to your config file: {PATH_TO_HMSS_CONFIG}")
            print("For reference, this is the default config file:")
            print(DEFAULT_HMSS_CONFIG)
            return None
        return result

    @staticmethod
    def write_defaults(overwrite: bool = False):
        """ Create the config file, as necessary. """
        if overwrite or not Path(PATH_TO_HMSS_CONFIG).exists():
            with open(PATH_TO_HMSS_CONFIG, "w") as config_file:
                json.dump(DEFAULT_HMSS_CONFIG, config_file, indent=JSON_INDENT)



"""
This code tests the RoyalRoyalReposBackup class.
"""

# Standard imports.
from pathlib import Path
from unittest.mock import Mock, patch

# Source imports.
from source.royal_repos_backup import RoyalReposBackup

# Local constants.
TEST_PATH_TO_HMSS_CONFIG = "test_hmss_config.json"
TEST_PATH_TO_LOG = "test.log"

###########
# TESTING #
###########

@patch("source.hmss_config.PATH_TO_HMSS_CONFIG", TEST_PATH_TO_HMSS_CONFIG)
@patch("source.royal_repos_backup.PATH_TO_LOG", TEST_PATH_TO_LOG)
def test_royal_repos_backup():
    """ Test that the class works as intended. """
    Path(TEST_PATH_TO_HMSS_CONFIG).unlink(missing_ok=True)
    Path(TEST_PATH_TO_LOG).unlink(missing_ok=True)
    backup_obj = RoyalReposBackup(human_interface=True)
    backup_obj = RoyalReposBackup()
    backup_obj._run_git_command = Mock(return_value=True)
    backup_obj.back_up_all()
    Path(TEST_PATH_TO_HMSS_CONFIG).unlink()
    Path(TEST_PATH_TO_LOG).unlink()



"""
This code defines a class which installs the various packages and repositories
required on this computer.
"""

# Standard imports.
import subprocess
from dataclasses import dataclass
from pathlib import Path

# Local imports.
from .hmss_config import HMSSConfig

# Local constants.
PATH_OBJ_TO_HERE = Path(__file__).parent
PATH_TO_INSTALL_SCRIPT_BASE = str(PATH_OBJ_TO_HERE/"sh"/"install_hmss_base.sh")
PATH_TO_INSTALL_SCRIPT_TEMP = str(Path.home()/"install_hmss_temp.sh")

##############
# MAIN CLASS #
##############

def install_hmss(*args, **kwargs) -> bool:
    """ Compress the class into a function. """
    installer_obj = HMSoftwareInstaller(*args, **kwargs)
    return installer_obj.run()

@dataclass
class HMSoftwareInstaller:
    """ The class in question. """
    human_interface: bool = False
    config: HMSSConfig|None = None

    def __post_init__(self):
        if not self.config:
            if self.human_interface:
                self.config = HMSSConfig.read_human()
            else:
                self.config = HMSSConfig.read_machine()

    def _write_install_script(self):
        """ Read the base file; make the replacements; write the new file. """
        with open(PATH_TO_INSTALL_SCRIPT_BASE, "r") as base_file:
            script = base_file.read()
        replacements = (
            (
                "%ESSENTIAL_APT_PACKAGES%",
                sh_list(self.config.essential_apt_packages)
            ), (
                "%INSTALL_CHROME%",
                sh_bool(self.config.install_chrome)
            ), (
                "%NON_ESSENTIAL_APT_PACKAGES%",
                sh_list(self.config.non_essential_apt_packages)
            ), (
                "%ROYAL_REPOS%",
                sh_list(self.config.royal_repos)
            ), (
                "%CLONE_METHOD%",
                self.config.clone_method
            ), (
                "%GIT_HOST%",
                self.config.git_host
            ), (
                "%GIT_ACCOUNT_NAME%",
                self.config.git_account_name
            ), (
                "%PATH_TO_WALLPAPER%",
                self.config.path_to_wallpaper_file
            )
        )
        for pair in replacements:
            script = script.replace(*pair)
        with open(PATH_TO_INSTALL_SCRIPT_TEMP, "w") as temp_file:
            temp_file.write(script)

    def _run_install_script(self) -> bool:
        """ Ronseal. """
        try:
            subprocess.run(["sh", PATH_TO_INSTALL_SCRIPT_TEMP], check=True)
        except subprocess.CalledProcessError:
            return False
        return True

    def _clean(self):
        """ Remove any temporary files which have served their purpose. """
        Path(PATH_TO_INSTALL_SCRIPT_TEMP).unlink(missing_ok=True)

    def run(self):
        """ Run the installation routine. """
        if not self.config:
            return False
        self._write_install_script()
        if self._run_install_script():
            self._clean()
            return True
        return False

################################
# HELPER CLASSES AND FUNCTIONS #
################################

class HMSSError(Exception):
    """ A custom exception. """

def sh_bool(py_bool: bool) -> str:
    """ Convert a Python boolean to its Shell Script equivalent. """
    if py_bool is True:
        return "true"
    if py_bool is False:
        return "false"
    raise HMSSError(f"Not a boolean: {py_bool}")

def sh_list(py_list: list|None) -> str:
    """ Convert a Python list to its Shell Script equivalent. """
    if not py_list:
        return ""
    result = " ".join(py_list)
    return result



"""
This code defines a function which installs a given list of PIP packages.
"""

# Standard imports.
import subprocess

#############
# FUNCTIONS #
#############

def install_dependency(package: str) -> bool:
    """
    Install a PIP package from a given package string, e.g. "pytest",
    "pylint>=2.12.2", etc.
    """
    try:
        subprocess.run(["pip", "install", package], check=True)
    except subprocess.CalledProcessError:
        return False
    return True

def install_dependencies(packages: list[str]) -> bool:
    """ As above, but for several packages. """
    for package in packages:
        local_result = \
            install_dependency(package)
        if not local_result:
            return False
    return True

def install_apt_package(
    package: str,
    yes: bool = True,
    raise_error: bool = True,
    quiet: bool = False
) -> bool:
    """ Obviously, this will only work in a Debian-based system. """
    args = ["sudo", "apt-get", "install"]
    if yes:
        args += ["--yes"]
    args += [package]
    if not quiet:
        print(f"I'm going to need superuser privileges to install {package}")
    try:
        subprocess.run(args, check=True)
    except subprocess.CalledProcessError:
        if raise_error:
            raise
        return False
    return True

def install_apt_packages(
    packages: list[str],
    quiet: bool = False,
    **kwargs
) -> bool:
    """ An iterative version of the above. """
    if not quiet:
        print(
            "I'm going to need superuser privileges to install %s"
            % ", ".join(packages)
        )
    for package in packages:
        if not install_apt_package(package, quiet=False, **kwargs):
            return False
    return True



"""
This code tests the HMSSConfig class.
"""

# Standard imports.
from pathlib import Path
from unittest.mock import Mock, patch

# Source imports.
from source.hm_software_installer import HMSoftwareInstaller

# Local constants.
TEST_PATH_TO_HMSS_CONFIG = "test_hmss_config.json"

###########
# TESTING #
###########

@patch("source.hmss_config.PATH_TO_HMSS_CONFIG", TEST_PATH_TO_HMSS_CONFIG)
def test_hm_software_installer():
    """ Test that the class works as intended. """
    Path(TEST_PATH_TO_HMSS_CONFIG).unlink(missing_ok=True)
    installer_obj = HMSoftwareInstaller(human_interface=True)
    assert not installer_obj.run()
    installer_obj = HMSoftwareInstaller()
    installer_obj._run_install_script = Mock(return_value=True)
    assert installer_obj.run()
    Path(TEST_PATH_TO_HMSS_CONFIG).unlink()



"""
This code gives us, amongst other things, the ability to import specific
classes and functions from the package directly, rather than having to go
through the modules in which they are defined.
"""

# Local imports.
from .continuous_integration import run_continuous_integration
from .hm_software_installer import HMSoftwareInstaller, install_hmss
from .install_dependencies import (
    install_dependency,
    install_dependencies,
    install_apt_package,
    install_apt_packages
)
from .misc import get_yes_no
